from __future__ import absolute_import
from datetime import datetime
from cloudperf.providers import aws_helpers

# static map until Amazon can provide the name in boto3 along with the region
# code...
region_map = {
    'Africa (Cape Town)': 'af-south-1',
    'Asia Pacific (Hong Kong)': 'ap-east-1',
    'Asia Pacific (Mumbai)': 'ap-south-1',
    'Asia Pacific (Osaka-Local)': 'ap-northeast-3',
    'Asia Pacific (Seoul)': 'ap-northeast-2',
    'Asia Pacific (Singapore)': 'ap-southeast-1',
    'Asia Pacific (Sydney)': 'ap-southeast-2',
    'Asia Pacific (Tokyo)': 'ap-northeast-1',
    'AWS GovCloud (US-East)': 'us-gov-east-1',
    'AWS GovCloud (US-West)': 'us-gov-west-1',
    'AWS GovCloud (US)': 'us-gov-west-1',
    'Canada (Central)': 'ca-central-1',
    'EU (Frankfurt)': 'eu-central-1',
    'EU (Ireland)': 'eu-west-1',
    'EU (London)': 'eu-west-2',
    'EU (Paris)': 'eu-west-3',
    'EU (Stockholm)': 'eu-north-1',
    'Middle East (Bahrain)': 'me-south-1',
    'South America (Sao Paulo)': 'sa-east-1',
    'US East (N. Virginia)': 'us-east-1',
    'US East (Ohio)': 'us-east-2',
    'US West (Los Angeles)': 'us-west-2-lax-1',
    'US West (N. California)': 'us-west-1',
    'US West (Oregon)': 'us-west-2'
}


class CloudProvider(object):
    provider = 'aws'
    filters = {'operatingSystem': 'Linux', 'preInstalledSw': 'NA',
               'licenseModel': 'No License required', 'capacitystatus': 'Used',
               'tenancy': 'Shared'
               }

    def get_prices(self, **filters):
        if not filters:
            filters = self.filters
        instances = aws_helpers.get_ec2_prices(**filters)
        # add a provider column
        instances['provider'] = self.provider

        return instances

    def get_performance(self, prices_df, perf_df=None, update=None, expire=None, tags=[], **filters):
        if not filters:
            filters = self.filters
        # only pass our records
        prices_df = prices_df[prices_df['provider'] == self.provider]
        if perf_df is not None:
            perf_df = perf_df[perf_df['provider'] == self.provider]
        instances = aws_helpers.get_ec2_performance(prices_df, perf_df, update, expire, tags, **filters)
        if instances.empty:
            return instances
        # add a provider column
        instances['provider'] = self.provider

        return instances

    def terminate_instances(self):
        aws_helpers.terminate_instances()



from __future__ import absolute_import
import os
import importlib
import pkgutil
import cloudperf.providers
import cachetools
import pandas as pd

prices_url = 'https://cloudperf-data.s3-us-west-2.amazonaws.com/prices.json.gz'
performance_url = 'https://cloudperf-data.s3-us-west-2.amazonaws.com/performance.json.gz'


def set_fail_on_exit():
    os.environ['FAIL_ON_EXIT'] = '1'


def fail_on_exit():
    if os.environ.get('FAIL_ON_EXIT'):
        return True
    return False


def sftp_write_file(sftp, name, contents, mode=0o755):
    f = sftp.open(name, 'w')
    f.write(contents)
    f.close()
    if mode is not None:
        sftp.chmod(name, mode)


class DictQuery(dict):
    def get(self, keys, default=None):
        val = None

        for key in keys:
            if val:
                if isinstance(val, list):
                    val = [v.get(key, default) if v else None for v in val]
                else:
                    try:
                        val = val.get(key, default)
                    except AttributeError:
                        return default
            else:
                val = dict.get(self, key, default)

            if val == default:
                break

        return val


@cachetools.cached(cache={})
def get_providers():
    prov_path = cloudperf.providers.__path__
    providers = []

    for _, name, _ in pkgutil.iter_modules(prov_path):
        m = importlib.import_module(name='.{}'.format(name), package='cloudperf.providers')
        if getattr(m, 'CloudProvider', None):
            providers.append(m.CloudProvider())
    return providers


def get_prices(prices=None, update=False):
    # if we got a stored file and update is True, merge the two by overwriting
    # old data with new (and leaving not updated old data intact)
    if prices and update:
        old = pd.read_json(prices, orient='records')
        new = pd.concat([cp.get_prices() for cp in get_providers()], ignore_index=True, sort=False)
        if new.empty:
            return old
        # update rows which have the same values in the following columns
        indices = ['provider', 'instanceType', 'region', 'spot', 'spot-az']
        return new.set_index(indices).combine_first(old.set_index(indices)).reset_index()
    if prices:
        return pd.read_json(prices, orient='records')
    return pd.concat([cp.get_prices() for cp in get_providers()], ignore_index=True, sort=False)


def args_cache_key(*args, **kw):
    args = list(args)
    for k, v in kw.items():
        if isinstance(v, list):
            v = tuple(v)
        try:
            hash(v)
        except Exception:
            continue
        args.append((k, v))
    return tuple(args)


def get_performance(prices=None, perf=None, update=False, expire=False, tags=[], maxcpu=False):
    # if we got a stored file and update is True, merge the two by overwriting
    # old data with new (and leaving not updated old data intact).
    # if expire is set only update old data if the expiry period is passed
    if perf and update:
        old = pd.read_json(perf, orient='records')
        new = pd.concat([cp.get_performance(get_prices(prices), old, update, expire, tags=tags) for cp in get_providers()],
                        ignore_index=True, sort=False)
        if new.empty:
            resdf = old
        else:
            # update rows which have the same values in the following columns
            indices = ['provider', 'instanceType', 'benchmark_id', 'benchmark_cpus']
            resdf = new.set_index(indices).combine_first(old.set_index(indices)).reset_index()
    elif perf:
        resdf = pd.read_json(perf, orient='records')
    else:
        resdf = pd.concat([cp.get_performance(get_prices(prices), tags=tags) for cp in get_providers()], ignore_index=True, sort=False)
    if maxcpu:
        return resdf.sort_values('benchmark_cpus', ascending=False).drop_duplicates(['instanceType', 'benchmark_id'])
    else:
        return resdf


def get_combined(prices=prices_url, perf=performance_url, maxcpu=False, spot_duration=None):
    prices_df = get_prices(prices=prices)
    perf_df = get_performance(prices=prices, perf=perf, maxcpu=maxcpu)
    combined_df = perf_df.merge(prices_df, how='left', on=['provider', 'instanceType'], suffixes=('', '_prices'))
    if spot_duration:
        duration_field = f'price_{spot_duration:.0f}h'
        if duration_field in combined_df:
            combined_df.loc[combined_df.spot, 'price'] = combined_df[duration_field]

    combined_df['perf/price/cpu'] = combined_df['benchmark_score']/combined_df['price']/combined_df['benchmark_cpus']
    combined_df['perf/price'] = combined_df['benchmark_score']/combined_df['price']

    return combined_df


def terminate_instances():
    for cp in get_providers():
        cp.terminate_instances()



from __future__ import absolute_import
from .core import get_prices, get_performance, get_combined, prices_url, performance_url, terminate_instances

__all__ = ['get_prices', 'get_performance', 'get_combined',
           'prices_url', 'performance_url', 'terminate_instances']
__version__ = '0.0.20'



from __future__ import absolute_import
import base64
import re
import sys
import json
import time
import threading
import logging
import functools
import collections
from logging import NullHandler
import copy
from datetime import datetime, date
from io import StringIO
from multiprocessing.pool import ThreadPool
import boto3
import cachetools
import requests
import paramiko
import pandas as pd
from dateutil import parser
from botocore.exceptions import ClientError
from cloudperf.benchmarks import benchmarks
from cloudperf.core import sftp_write_file, DictQuery, set_fail_on_exit


session = boto3.session.Session()
logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())

# get current defined-duration spot prices from here
spot_js = 'https://spot-price.s3.amazonaws.com/spotblocks-generic.js'

# blacklist instances (prefixes) until a given date (preview etc)
instance_blacklist = {'c6g': date(2020, 4, 1),
                      'm6g': date(2020, 2, 1),
                      'r6g': date(2020, 4, 1),
                      'cc2.8xlarge': date(9999, 1, 1),
                      }

# Self-destruct the machine after 2 hours
userdata_script = """#!/bin/sh
shutdown +120"""
ssh_keyname = 'batch'
ssh_user = 'ec2-user'
ssh_get_conn_timeout = 600
ssh_exec_timeout = 600
ec2_specs = {'KeyName': ssh_keyname, 'SecurityGroups': ['tech-ssh'],
             'MaxCount': 1, 'MinCount': 1, 'Monitoring': {'Enabled': False},
             'InstanceInitiatedShutdownBehavior': 'terminate',
             'UserData': userdata_script,
             'TagSpecifications': [{'ResourceType': 'instance',
                                    'Tags': [{'Value': 'cloudperf', 'Key': 'Application'}]},
                                   {'ResourceType': 'volume',
                                    'Tags': [{'Value': 'cloudperf', 'Key': 'Application'}]}]}

instance_init_script = """#!/bin/sh
sudo systemctl stop acpid chronyd crond ecs postfix
sudo curl -L https://github.com/docker/compose/releases/download/1.23.2/docker-compose-`uname -s`-`uname -m` -o /usr/local/bin/docker-compose
sudo chmod +x /usr/local/bin/docker-compose
"""


def boto3_paginate(method, **kwargs):
    client = method.__self__
    paginator = client.get_paginator(method.__name__)
    for page in paginator.paginate(**kwargs).result_key_iters():
        for result in page:
            yield result


def ping_region(region, latencies, lock):
    st = time.time()
    try:
        requests.get('http://ec2.{}.amazonaws.com/ping'.format(region), timeout=1)
    except Exception:
        return
    with lock:
        latencies[region] = time.time()-st


def aws_ping(regions):
    latencies = {}
    lock = threading.Lock()
    threads = []
    for region in regions:
        t = threading.Thread(target=ping_region, args=(region, latencies, lock))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()
    return latencies


@cachetools.cached(cache={})
def aws_get_parameter(name):
    ssm = session.client('ssm', region_name=aws_get_region())
    res = ssm.get_parameter(Name=name, WithDecryption=True)
    try:
        return json.loads(res['Parameter']['Value'])
    except Exception:
        return res['Parameter']['Value']


def aws_get_cpu_arch(instance):
    # XXX: maybe in the future Amazon will indicate the exact CPU architecture,
    # but until that...
    physproc = DictQuery(instance).get(['product', 'attributes', 'physicalProcessor'], '').lower()
    procarch = DictQuery(instance).get(['product', 'attributes', 'processorArchitecture'], '').lower()
    instance_type = DictQuery(instance).get(['product', 'attributes', 'instanceType'], '').lower()
    if re.match('^a[0-9]+\.', instance_type) or re.search('aws\s+(graviton.*|)\s*processor', physproc):
        # try to find arm instances
        return 'arm64'

    return 'x86_64'


def aws_get_region():
    region = boto3.session.Session().region_name
    if region:
        return region
    try:
        r = requests.get(
            'http://169.254.169.254/latest/dynamic/instance-identity/document')
        return r.json().get('region')
    except Exception:
        return None


def aws_newest_image(imgs):
    latest = None

    for image in imgs:
        if not latest:
            latest = image
            continue

        if parser.parse(image['CreationDate']) > parser.parse(latest['CreationDate']):
            latest = image

    return latest


@cachetools.cached(cache={})
def aws_get_latest_ami(name='amzn2-ami-ecs-hvm*ebs', arch='x86_64'):
    ec2 = session.client('ec2', region_name=aws_get_region())

    filters = [
        {'Name': 'name', 'Values': [name]},
        {'Name': 'description', 'Values': ['Amazon Linux AMI*']},
        {'Name': 'architecture', 'Values': [arch]},
        {'Name': 'owner-alias', 'Values': ['amazon']},
        {'Name': 'state', 'Values': ['available']},
        {'Name': 'root-device-type', 'Values': ['ebs']},
        {'Name': 'virtualization-type', 'Values': ['hvm']},
        {'Name': 'image-type', 'Values': ['machine']}
    ]

    response = ec2.describe_images(Owners=['amazon'], Filters=filters)
    return aws_newest_image(response['Images'])


def get_running_ec2_instances(filters=[]):
    ec2 = session.client('ec2', region_name=aws_get_region())
    response = ec2.describe_instances(Filters=filters)
    instances = []
    for reservation in response["Reservations"]:
        for instance in reservation["Instances"]:
            if DictQuery(instance).get(['State', 'Name']) == 'running':
                instances.append(instance)
    return instances


def terminate_instances():
    filter = [{'Name': 'tag:Application', 'Values': ['cloudperf']}]
    tag = {'Key': 'Application', 'Value': 'cloudperf'}
    ec2 = session.client('ec2', region_name=aws_get_region())
    for instance in get_running_ec2_instances(filter):
        # although we filter for our tag, an error in this would cause the
        # termination of other machines, so do a manual filtering here as well
        if tag not in instance['Tags']:
            continue
        logger.info("Terminating instance {}".format(instance['InstanceId']))
        ec2.terminate_instances(InstanceIds=[instance['InstanceId']])


@cachetools.cached(cache={}, key=tuple)
def closest_regions(regions):
    latencies = aws_ping(regions)
    regions.sort(key=lambda k: latencies.get(k, 9999))
    return regions


def aws_format_memory(memory):
    return "{:,g} GiB".format(float(memory))


def aws_parse_memory(memory):
    # currently only GiBs are returned, so we don't need to take unit into account
    number, unit = memory.split()
    return float(number.replace(',', ''))


@cachetools.cached(cache={})
def get_region():
    region = boto3.session.Session().region_name
    if region:
        return region
    try:
        r = requests.get(
            'http://169.254.169.254/latest/dynamic/instance-identity/document')
        return r.json().get('region')
    except Exception:
        return None


@cachetools.cached(cache={})
def get_regions():
    client = session.client('ec2')
    return [region['RegionName'] for region in client.describe_regions()['Regions']]


@cachetools.cached(cache={})
def get_ec2_instances(**filter_opts):
    """Get AWS instances according to the given filter criteria

    Args:
        any Field:Value pair which the AWS API accepts.
        Example from a c5.4xlarge instance:
        {'capacitystatus': 'Used',
         'clockSpeed': '3.0 Ghz',
         'currentGeneration': 'Yes',
         'dedicatedEbsThroughput': 'Upto 2250 Mbps',
         'ecu': '68',
         'enhancedNetworkingSupported': 'Yes',
         'instanceFamily': 'Compute optimized',
         'instanceType': 'c5.4xlarge',
         'licenseModel': 'No License required',
         'location': 'US West (Oregon)',
         'locationType': 'AWS Region',
         'memory': '32 GiB',
         'networkPerformance': 'Up to 10 Gigabit',
         'normalizationSizeFactor': '32',
         'operatingSystem': 'Linux',
         'operation': 'RunInstances:0004',
         'physicalProcessor': 'Intel Xeon Platinum 8124M',
         'preInstalledSw': 'SQL Std',
         'processorArchitecture': '64-bit',
         'processorFeatures': 'Intel AVX, Intel AVX2, Intel AVX512, Intel Turbo',
         'servicecode': 'AmazonEC2',
         'servicename': 'Amazon Elastic Compute Cloud',
         'storage': 'EBS only',
         'tenancy': 'Host',
         'usagetype': 'USW2-HostBoxUsage:c5.4xlarge',
         'vcpu': '16'}

    Returns:
        type: dict of AWS product descriptions

    """
    filters = [{'Type': 'TERM_MATCH', 'Field': k, 'Value': v}
               for k, v in filter_opts.items()]

    # currently the pricing API is limited to some regions, so don't waste time
    # on trying to access it on others one by one
    # regions = get_regions()
    regions = ['us-east-1', 'ap-south-1']
    for region in closest_regions(regions):
        pricing = session.client('pricing', region_name=region)
        instances = []
        for data in boto3_paginate(pricing.get_products, ServiceCode='AmazonEC2', Filters=filters, MaxResults=100):
            pd = json.loads(data)
            instances.append(pd)
        break
    return instances


def get_ec2_defined_duration_prices():
    """Get AWS defined-duration prices from the web. Currently there's no
    API for this, so we'll use the JavaScript used by the public EC2 spot
    pricing page: https://aws.amazon.com/ec2/spot/pricing/
    We deliberately lack error handling here, so we can detect any failures in
    the parsing process.
    """

    r = requests.get(spot_js)
    # this is JavaScript, so we have to parse data out from it
    js = r.text
    data = json.loads(js[js.find('{'):js.rfind('}')+1])

    # create a structure of region:instance_type:duration prices similar to this:
    # {'us-west-2': {'g4dn.xlarge': {1: 0.307,
    #                                2: 0.335,
    #                                3: 0.349,
    #                                4: 0.363,
    #                                5: 0.377,
    #                                6: 0.391}}}
    # the keys in the instance's dictionary is the duration in hours, where
    # we fill up the missing durations in the input data with a linear estimation
    block_data = collections.defaultdict(lambda: collections.defaultdict(dict))
    for region_data in data['config']['regions']:
        region = region_data['region']
        for instance_data in region_data['instanceTypes']:
            for instance in instance_data['sizes']:
                instance_type = instance['size']
                for duration_data in instance['valueColumns']:
                    # name is like '1 hour' or '6 hours'
                    m = re.search('[0-9]+', duration_data['name'])
                    if not m:
                        continue
                    duration = int(m.group(0))
                    block_data[region][instance_type][duration] = float(duration_data['prices']['USD'])
        # fill up gaps in defined durations by estimating the hourly price
        for instance_type, instance_data in block_data[region].items():
            min_duration = min(instance_data.keys())
            max_duration = max(instance_data.keys())
            min_price = instance_data[min_duration]
            max_price = instance_data[max_duration]
            step = (max_price-min_price)/max_duration
            for i in range(min_duration, max_duration):
                if i in instance_data:
                    continue
                # round to 3 digits precision
                instance_data[i] = round(min_price+step*i, 3)

    return block_data


def get_ec2_prices(**filter_opts):
    """Get AWS instance prices according to the given filter criteria

    Args:
        get_instance_types arguments

    Returns:
        DataFrame with instance attributes and pricing

    """
    from cloudperf.providers.aws import region_map
    prices = []
    params = {}

    for data in get_ec2_instances(**filter_opts):
        try:
            instance_type = data['product']['attributes']['instanceType']
            price = float(list(list(data['terms']['OnDemand'].values())[
                          0]['priceDimensions'].values())[0]['pricePerUnit']['USD'])
        except Exception:
            continue
        if price == 0:
            continue
        if data['product']['attributes']['memory'] == 'NA' or \
                data['product']['attributes']['vcpu'] == 'NA':
            # skip these
            continue
        vcpu = int(data['product']['attributes']['vcpu'])
        memory = aws_parse_memory(data['product']['attributes']['memory'])
        try:
            region = region_map[data['product']['attributes']['location']]
        except KeyError:
            print(f""""{data['product']['attributes']['location']}" is missing from region_map, """
                  'please update (for eg. from https://aws.amazon.com/ec2/pricing/on-demand/, '
                  'inspecting the region dropdown, or https://docs.aws.amazon.com/general/latest/gr/rande.html)')
            sys.exit(1)
        params[instance_type] = data['product']['attributes']
        params[instance_type].update({'vcpu': vcpu, 'memory': memory, 'region': region,
                                      'cpu_arch': aws_get_cpu_arch(data),
                                      'date': datetime.now()})
        d = {'price': price, 'spot': False, 'spot-az': None}
        d.update(params[instance_type])
        prices.append(d)

    if not prices:
        # we couldn't find any matching instances
        return prices

    # get actual defined-duration spot prices from the web, until the pricing
    # API supports these...
    block_prices = get_ec2_defined_duration_prices()

    for region in get_regions():
        ec2 = session.client('ec2', region_name=region)
        for data in boto3_paginate(ec2.describe_spot_price_history, InstanceTypes=list(params.keys()),
                                   MaxResults=100, ProductDescriptions=['Linux/UNIX (Amazon VPC)'], StartTime=datetime.now()):
            instance_type = data['InstanceType']
            d = copy.deepcopy(params[instance_type])
            d.update({'price': float(data['SpotPrice']), 'spot': True, 'spot-az': data['AvailabilityZone'], 'region': region})
            for duration, price in DictQuery(block_prices).get([region, instance_type], {}).items():
                # add spot blocked duration prices, if any
                d.update({f'price_{duration}h': price})
            prices.append(d)

    return pd.DataFrame.from_dict(prices)


def get_ssh_connection(instance, user, pkey, timeout):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    start = time.time()
    while start+timeout > time.time():
        try:
            ssh.connect(instance['PrivateIpAddress'], username=user, pkey=pkey, timeout=10, auth_timeout=10)
            break
        except Exception as e:
            logger.info("Couldn't connect: {}, retrying for {:.0f}s".format(e, start+timeout-time.time()))
            time.sleep(5)
    else:
        return None
    return ssh


def log_exception(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception:
            err = "Exception in {}".format(function.__name__)
            logging.exception(err)
    return wrapper


@log_exception
def run_benchmarks(args):
    threading.current_thread().name = 'run_bench'
    ami, instance, tags, benchmarks_to_run = args
    specs = copy.deepcopy(ec2_specs)
    # extend tagspecs with user specified tags
    tagspecs = [{'Key': k, 'Value': v} for k, v in tags]
    for i in specs.get('TagSpecifications', []):
        i['Tags'].extend(tagspecs)
    bdmap = ami['BlockDeviceMappings']
    try:
        # You cannot specify the encrypted flag if specifying a snapshot id in a block device mapping.
        del bdmap[0]['Ebs']['Encrypted']
    except Exception:
        pass
    specs.update({'BlockDeviceMappings': bdmap,
                  'ImageId': ami['ImageId'], 'InstanceType': instance.instanceType})
    # add unlimited cpu credits on burstable type instances, so these won't affect
    # benchmark results
    if re.match('^t[0-9]+\.', instance.instanceType):
        specs.update({'CreditSpecification': {'CpuCredits': 'unlimited'}})
    spotspecs = copy.deepcopy(specs)
    spotspecs.update({'InstanceMarketOptions': {'MarketType': 'spot',
                                                'SpotOptions': {
                                                    'MaxPrice': str(instance.price),
                                                    'SpotInstanceType': 'one-time',
                                                    'InstanceInterruptionBehavior': 'terminate'
                                                    }
                                                }})
    # start with a spot instance
    create_specs = spotspecs
    retcount = 0
    ec2_inst = None
    ec2 = session.client('ec2', region_name=aws_get_region())
    while retcount < 16:
        try:
            ec2_inst = ec2.run_instances(**create_specs)['Instances'][0]
            break
        except ClientError as e:
            # retry on request limit exceeded
            if e.response['Error']['Code'] == 'RequestLimitExceeded':
                logger.info("Request limit for {}: {}, retry #{}".format(instance.instanceType,
                                                                           e.response['Error']['Message'], retcount))
                time.sleep(1.2**retcount)
                retcount += 1
                continue

            if e.response['Error']['Code'] == 'InsufficientInstanceCapacity':
                logger.info('Insufficient capacity for {}: {}'.format(
                    instance.instanceType, e))
                if create_specs == specs:
                    # if it was on-demand retry until the counter expires
                    time.sleep(1.2**retcount)
                    retcount += 1
                else:
                    # retry with on demand if this was with spot
                    create_specs = specs
                    retcount = 0
                continue

            if e.response['Error']['Code'] == 'SpotMaxPriceTooLow':
                try:
                    # the actual spot price is the second, extract it
                    sp = re.findall('[0-9]+\.[0-9]+',
                                    e.response['Error']['Message'])[1]
                    logger.info(
                        "Spot price too low spotmax:{}, current price:{}".format(instance.price, sp))
                except Exception:
                    logger.info("Spot price too low for {}, {}".format(
                        instance.instanceType, e.response['Error']['Message']))
                # retry with on demand
                create_specs = specs
                retcount = 0
                continue

            if e.response['Error']['Code'] == 'MissingParameter':
                # this can't be fixed, exit
                logger.error("Missing parameter while creating {}: {}".format(
                    instance.instanceType, e))
                set_fail_on_exit()
                break

            if e.response['Error']['Code'] == 'InvalidParameterValue':
                # certain instances are not allowed to be created
                logger.error("Error starting instance {}: {}".format(
                    instance.instanceType, e.response['Error']['Message']))
                if retcount == 0:
                    # retry with on demand
                    create_specs = specs
                    retcount += 1
                    continue
                set_fail_on_exit()
                break

            if e.response['Error']['Code'] == 'Unsupported':
                # certain instances are not allowed to be created
                logger.error("Unsupported instance {}: {}, specs: {}".format(
                    instance.instanceType, e.response['Error']['Message'],
                    base64.b64encode(json.dumps(create_specs).encode('utf-8'))))
                break

            if e.response['Error']['Code'] == 'InstanceCreditSpecification.NotSupported':
                # remove unlimited credit and try again
                logger.error("{} doesn't support unlimited credits: {}".format(
                    instance.instanceType, e.response['Error']['Message']))
                if 'CreditSpecification' in create_specs:
                    del create_specs['CreditSpecification']
                    retcount += 1
                    continue
                else:
                    break

            logger.error("Other error while creating {}: {}, code: {}".format(
                instance.instanceType, e, DictQuery(e.response).get(['Error', 'Code'])))
            time.sleep(1.2**retcount)
            retcount += 1

        except Exception as e:
            logger.error("Other exception while creating {}: {}".format(
                instance.instanceType, e))
            time.sleep(1.2**retcount)
            retcount += 1

    if not ec2_inst:
        return None

    instance_id = ec2_inst['InstanceId']
    threading.current_thread().name = instance_id

    logger.info(
        "Waiting for instance {} to be ready".format(instance.instanceType))
    # wait for the instance
    try:
        waiter = ec2.get_waiter('instance_running')
        waiter.wait(InstanceIds=[instance_id], WaiterConfig={
            # wait for up to 30 minutes
            'Delay': 15,
            'MaxAttempts': 120
            })
    except Exception:
        logger.exception(
            'Waiter failed for {}'.format(instance.instanceType))

    # give 5 secs before trying ssh
    time.sleep(5)
    pkey = paramiko.RSAKey.from_private_key(
        StringIO(aws_get_parameter('/ssh_keys/{}'.format(ssh_keyname))))
    ssh = get_ssh_connection(ec2_inst, ssh_user, pkey, ssh_get_conn_timeout)
    if ssh is None:
        logger.error("Couldn't open an ssh connection, terminating instance")
        ec2.terminate_instances(InstanceIds=[instance_id])
        return None

    sftp = ssh.open_sftp()

    # write init_script
    for i in range(4):
        try:
            sftp_write_file(sftp, 'init_script', instance_init_script)
            break
        except Exception:
            logger.exception("Failed to write init_script, try #{}".format(i))
            continue
    # try stop all unnecessary services in order to provide a more reliable result
    for i in range(4):
        logger.info("Trying to execute init_script, try #{}".format(i))
        stdin, stdout, stderr = ssh.exec_command("./init_script", timeout=ssh_exec_timeout)
        if stdout.channel.recv_exit_status() == 0:
            break
        time.sleep(5)
    else:
        logger.error("Couldn't execute init_script: {}, {}".format(
            stdout.read(), stderr.read()))
        ec2.terminate_instances(InstanceIds=[instance_id])
        return None

    # give some more time for the machine to be ready and to settle down
    time.sleep(20)

    results = []
    try:
        for name, bench_data in benchmarks_to_run.items():
            threading.current_thread().name = '{}/{}'.format(instance_id, name)
            docker_img = bench_data['images'].get(instance.cpu_arch, None)
            if not docker_img:
                logger.error("Couldn't find docker image for {}/{}".format(name, instance.cpu_arch))
                continue
            # write files for the benchmark
            for name, contents in bench_data['images'].get('files', {}):
                sftp_write_file(sftp, name, contents)

            # docker pull and wait some time
            for i in range(4):
                logger.info("Docker pull, try #{}".format(i))
                stdin, stdout, stderr = ssh.exec_command("docker pull {}; sync; sleep 10".format(docker_img), timeout=ssh_exec_timeout)
                if stdout.channel.recv_exit_status() == 0:
                    break
                time.sleep(5)
            else:
                logger.error("Couldn't pull docker image {}, {}".format(
                    stdout.read(), stderr.read()))
                continue

            if 'composefile' in bench_data:
                sftp_write_file(sftp, 'docker-compose.yml', bench_data['composefile'], 0o644)
                # start docker compose
                stdin, stdout, stderr = ssh.exec_command("docker-compose up -d", timeout=ssh_exec_timeout)
                if stdout.channel.recv_exit_status() != 0:
                    logger.error("Couldn't start docker compose {}, {}".format(
                        stdout.read(), stderr.read()))
                    continue

                if 'after_compose_up' in bench_data:
                    sftp_write_file(sftp, 'after_compose_up', bench_data['after_compose_up'])
                    stdin, stdout, stderr = ssh.exec_command("./after_compose_up", timeout=ssh_exec_timeout)
                    if stdout.channel.recv_exit_status() != 0:
                        logger.error("Couldn't start after_compose_up script {}, {}".format(
                            stdout.read(), stderr.read()))
                        continue

            if 'cpus' in bench_data and bench_data['cpus']:
                cpulist = bench_data['cpus']
            else:
                cpulist = range(1, instance.vcpu+1)

            # default options if missing
            docker_opts = bench_data.get('docker_opts', '--network none')
            for i in cpulist:
                ssh.exec_command("sync", timeout=ssh_exec_timeout)
                dcmd = bench_data['cmd'].format(numcpu=i)
                if 'timeout' in bench_data:
                    timeout_cmd = 'timeout -k {} {} '.format(bench_data['timeout']+5, bench_data['timeout'])
                else:
                    timeout_cmd = ''
                cmd = '{}docker run --rm {} {} {}'.format(timeout_cmd, docker_opts, docker_img, dcmd)
                scores = []
                for it in range(bench_data.get('iterations', 3)):
                    logger.info("Running command: {}, iter: #{}".format(cmd, it))
                    stdin, stdout, stderr = ssh.exec_command(cmd, timeout=ssh_exec_timeout)
                    ec = stdout.channel.recv_exit_status()
                    stdo = stdout.read()
                    stdrr = stderr.read()
                    if ec == 0:
                        try:
                            scores.append(float(stdo))
                        except Exception:
                            logger.info(
                                "Couldn't parse output: {}".format(stdo))
                            scores.append(None)
                    else:
                        logger.info("Non-zero exit code {}, {}, {}".format(ec, stdo, stdrr))
                aggr_f = bench_data.get('score_aggregation', max)
                try:
                    score = aggr_f(scores)
                except Exception:
                    score = None
                results.append({'instanceType': instance.instanceType,
                                'benchmark_cpus': i, 'benchmark_score': score, 'benchmark_id': name,
                                'benchmark_name': bench_data.get('name'),
                                'benchmark_cmd': cmd, 'benchmark_program': bench_data.get('program'),
                                'date': datetime.now()})
            if 'composefile' in bench_data:
                stdin, stdout, stderr = ssh.exec_command("docker-compose down -v", timeout=ssh_exec_timeout)
                if stdout.channel.recv_exit_status() != 0:
                    logger.error("Couldn't stop docker compose: {}, {}".format(
                        stdout.read(), stderr.read()))
                    continue
                if 'after_compose_down' in bench_data:
                    sftp_write_file(sftp, 'after_compose_down', bench_data['after_compose_down'])
                    stdin, stdout, stderr = ssh.exec_command("./after_compose_down", timeout=ssh_exec_timeout)
                    if stdout.channel.recv_exit_status() != 0:
                        logger.error("Couldn't start after_compose_down script: {}, {}".format(
                            stdout.read(), stderr.read()))
                        continue
    except Exception:
        logger.exception("Error while executing benchmarks")

    logger.info("Finished with instance, terminating")
    ec2.terminate_instances(InstanceIds=[instance_id])
    if results:
        return pd.DataFrame.from_dict(results)
    else:
        return None


def get_benchmarks_to_run(instance, perf_df, expire):
    my_benchmarks = copy.deepcopy(benchmarks)
    # filter the incoming perf data only to our instance type
    perf_df = perf_df[perf_df['instanceType'] == instance.instanceType][['instanceType', 'benchmark_id', 'date']].drop_duplicates()
    for idx, row in perf_df.iterrows():
        if (datetime.now() - row.date).seconds >= expire:
            # leave the benchmark if it's not yet expired ...
            continue
        # ... and drop, if it is
        my_benchmarks.pop(row.benchmark_id, None)

    return my_benchmarks


def is_blacklisted(instance):
    for prefix, dt in instance_blacklist.items():
        if instance.startswith(prefix) and datetime.now().date() <= dt:
            return True
    return False


def get_ec2_performance(prices_df, perf_df=None, update=None, expire=None, tags=[], **filter_opts):
    # drop spot instances
    prices_df = prices_df.drop(prices_df[prices_df.spot == True].index)
    # remove duplicate instances, so we'll have a list of all on-demand instances
    prices_df = prices_df.drop_duplicates(subset='instanceType')

    bench_args = []
    for instance in prices_df.itertuples():
        if is_blacklisted(instance.instanceType):
            logger.info("Skipping blacklisted instance: {}".format(instance.instanceType))
            continue
        ami = aws_get_latest_ami(arch=instance.cpu_arch)
        if perf_df is not None and update:
            benchmarks_to_run = get_benchmarks_to_run(instance, perf_df, expire)
        else:
            benchmarks_to_run = benchmarks

        if not benchmarks_to_run:
            logger.info("Skipping already benchmarked instance: {}".format(instance.instanceType))
            # leave this instance out if there is no benchmark to run
            continue
        ami = aws_get_latest_ami(arch=instance.cpu_arch)
        bench_args.append([ami, instance, tags, benchmarks_to_run])
    if bench_args:
        pool = ThreadPool(4)
        results = [res for res in pool.map(run_benchmarks, bench_args) if res is not None]
        if results:
            return pd.concat(results, ignore_index=True, sort=False)
    return pd.DataFrame({})



import io
import re
import setuptools

with io.open("README.md", "rt", encoding="utf8") as f:
    long_description = f.read()

with io.open("cloudperf/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r"__version__ = \'(.*?)\'", f.read()).group(1)

setuptools.setup(
    name="cloudperf",
    version=version,
    author="NAGY, Attila",
    author_email="nagy.attila@gmail.com",
    description="Relative performance index for cloud services",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bra-fsn/cloudperf",
    packages=setuptools.find_packages(),
    scripts=['bin/cloudperf'],
    install_requires=['cachetools', 'click>=7.0', 'boto3>=1.9.61', 'pandas', 'requests',
                      'python-dateutil', 'paramiko', 'pytimeparse'],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# This is a python file, defining benchmarks to run on each cloud instances.
# Each benchmark must have a unique name as the dictionary key, which should
# contain a dictionary for each benchmarks, with the following keys:
#   - `program`: the name of the benchmark program (informative, not used for execution)
#   - `name`: the verbose name of the benchmark
#   - `images`: another, architecture-name keyed dictionary, which defines a docker image
#               for each architecture
#   - `files`: a dictonary of file names and file contents (string) to write to
#              the host, like: {'setup_db.sh': crdb_setup},`
#   - `cmd`: the command to be passed to the docker image. Variable expansion
#            is available here with the following variables:
#               - `numcpu`: the excercised number of CPUs (if the benchmark can
#                           utilize more CPUs, it can be limited to that number).
#                           Without any other settings, the command will be invoked
#                           once for each CPUs, with an increasing number in `numcpu`
#            The variable expansion works by using `{numcpu}` in the command.
#            Due to this, `{}` must be escaped as `{{}}`.
#            The command must output a single number, parseable with python's
#            `float`, which will be the benchmark's score.
#            If the output can't be parsed with `float`, a `None` will be inserted,
#            so if the benchmark couldn't run on that instance (for example because
#            it doesn't have the needed amount of memory), we won't re-run the
#            benchmark forever, only after the expiration period is over.
#   - `timeout`: the cmd will be started with `timeout n`. This helps preventing a
#           stuck script to run forever.
#   - `cpus`: if not specified, the command will be run once for each CPU, having
#             the actual number of CPUs in the `numcpu` variable. For a 8 CPU
#             machine it means 8 runs with `numcpu` being 1,2,3,4,5,6,7,8.
#             `cpus` can be a python list, eg: 'cpus': [1,2,4,8,16]
#   - `iterations`: the benchmark will be runned this many times (for each CPUs)
#                   in order to have more measure points. Default: 3.
#   - `score_aggregation`: this python function will be used to aggregating the
#                          scores from the above iterations. The scores will be
#                          passed as a list of floats. Default: max

crdb_tag = 'v2.1.5'
stress_ng_tag = 'latest'

crdb_compose_yml = """version: '3'

services:
  db-0:
    image: cockroachdb/cockroach:{tag}
    command: start --insecure
    expose:
     - "8080"
     - "26257"
    ports:
     - "26257:26257"
     - "8080:8080"
    networks:
     - roachnet
    volumes:
     - ./data/db-0:/cockroach/cockroach-data
  db-1:
    image: cockroachdb/cockroach:{tag}
    command: start --insecure --join=db-0 --join=db-2
    networks:
     - roachnet
    volumes:
     - ./data/db-1:/cockroach/cockroach-data
  db-2:
    image: cockroachdb/cockroach:{tag}
    command: start --insecure --join=db-0 --join=db-1
    networks:
     - roachnet
    volumes:
     - ./data/db-2:/cockroach/cockroach-data
  # db-init:
  #  image: cockroachdb/cockroach:{tag}
  #  networks:
  #   - roachnet
  #  volumes:
  #    - ./setup_db.sh:/setup_db.sh
  #  entrypoint: "/bin/bash"
  #  command: /setup_db.sh
networks:
  roachnet:
""".format(tag=crdb_tag)

crdb_sysbench_common_opts = "--db-driver=pgsql --oltp-table-size=100000 --oltp-tables-count=24 " \
    "--pgsql-host=db-1 --pgsql-port=26257 --pgsql-user=root --pgsql-db=cloudperf"

crdb_compose_up = """#!/bin/sh
docker run --rm --network ec2-user_roachnet cockroachdb/cockroach:{} sql --host db-1 --insecure -e "CREATE DATABASE cloudperf;"
docker run --rm --network ec2-user_roachnet severalnines/sysbench sysbench --threads=1 {} /usr/share/sysbench/tests/include/oltp_legacy/parallel_prepare.lua prepare
""".format(crdb_tag, crdb_sysbench_common_opts)

crdb_compose_down = """#!/bin/sh
sudo rm -rf data
"""


mysql_compose_yml = """version: '3'

services:
  mysql:
    image: mysql/mysql-server:{}
    environment:
     - MYSQL_ROOT_PASSWORD=password
     - MYSQL_ROOT_HOST=%
    expose:
     - "3306"
    ports:
     - "3306:3306"
    networks:
     - mysql
    volumes:
     - ./data/mysql:/var/lib/mysql
networks:
  mysql:
"""

mysql_sysbench_common_opts = "--db-driver=mysql --oltp-table-size=100000 --oltp-tables-count=24 " \
    "--mysql-host=mysql --mysql-user=root --mysql-db=cloudperf"
mysql_compose_up = """#!/bin/sh
docker run --rm --network ec2-user_mysql mysql/mysql-server sh -c 'echo "create database cloudperf" | mysql -u root -ppassword -h mysql'
docker run --rm --network ec2-user_mysql severalnines/sysbench sysbench --threads=1 {} /usr/share/sysbench/tests/include/oltp_legacy/parallel_prepare.lua prepare
""".format(mysql_sysbench_common_opts)
mysql_compose_down = """#!/bin/sh
sudo rm -rf data
"""

benchmarks = {
    # 'sysbench:oltp:cockroachdb:{}'.format(crdb_tag): {'program': 'cockroachdb',
    #                                                   'name': 'stress-ng matrixprod',
    #                                                   'composefile': crdb_compose_yml,
    #                                                   'after_compose_up': crdb_compose_up,
    #                                                   'after_compose_down': crdb_compose_down,
    #                                                   # 'files': {'setup_db.sh': crdb_setup},
    #                                                   'docker_opts': '--network ec2-user_roachnet',
    #                                                   'cmd': "sysbench --threads={numcpu} " + crdb_sysbench_common_opts + " /usr/share/sysbench/tests/include/oltp_legacy/oltp.lua run | fgrep 'queries:' | egrep -o '[0-9.]+ per sec' | awk '{{print $1}}'",
    #                                                   'images': {'x86_64': 'severalnines/sysbench'}
    #                                                   },
    # 'sysbench:oltp:mysql:5.5': {'program': 'mysql',
    #                          'name': 'stress-ng matrixprod',
    #                          'composefile': mysql_compose_yml.format('5.5'),
    #                          'after_compose_up': mysql_compose_up,
    #                          'after_compose_down': mysql_compose_down,
    #                          # 'files': {'setup_db.sh': crdb_setup},
    #                          'docker_opts': '--network ec2-user_mysql',
    #                          'cmd': "sysbench --threads={numcpu} " + mysql_sysbench_common_opts + " /usr/share/sysbench/tests/include/oltp_legacy/oltp.lua run | fgrep 'queries:' | egrep -o '[0-9.]+ per sec' | awk '{{print $1}}'",
    #                          'images': {'x86_64': 'severalnines/sysbench'}
    #                          },
    # 'sysbench:cpu': {'program': 'sysbench',
    #                  'name': 'sysbench CPU performance test',
    #                          'cmd': "sysbench cpu --max-time=5 --num-threads={numcpu} run | fgrep 'events per second' | egrep -o '[0-9.]+'",
    #                          'images': {'x86_64': 'severalnines/sysbench'}
    #                  },
    'stress-ng:hdd_rndwr_512': {'program': 'stress-ng',
                                'name': 'Random RW 512 bytes O_DIRECT, O_DSYNC',
                                'cmd': "--hdd {numcpu} --hdd-opts direct,dsync,rd-rnd,wr-rnd --hdd-write-size 512 -t 5 --metrics 2>&1 | tail -1 | awk '{{print $9}}'",
                                'cpus': [256],
                                'timeout': 10,
                                'images': {'x86_64': 'brafsn/stress-ng-x86_64:{}'.format(stress_ng_tag),
                                           'arm64': 'brafsn/stress-ng-arm64:{}'.format(stress_ng_tag)},
                                },
    'stress-ng:hdd_rndwr_4k': {'program': 'stress-ng',
                               'name': 'Random RW 4k O_DIRECT, O_DSYNC',
                               'cmd': "--hdd {numcpu} --hdd-opts direct,dsync,rd-rnd,wr-rnd --hdd-write-size 4k -t 5 --metrics 2>&1 | tail -1 | awk '{{print $9}}'",
                               'cpus': [256],
                               'timeout': 10,
                               'images': {'x86_64': 'brafsn/stress-ng-x86_64:{}'.format(stress_ng_tag),
                                          'arm64': 'brafsn/stress-ng-arm64:{}'.format(stress_ng_tag)},
                               },
    'stress-ng:crc16': {'program': 'stress-ng',
                        'name': 'compute 1024 rounds of CCITT CRC16 on random data',
                        'cmd': "--cpu {numcpu} --cpu-method crc16 -t 5 --metrics 2>&1 | tail -1 | awk '{{print $9}}'",
                        'timeout': 10,
                        'images': {'x86_64': 'brafsn/stress-ng-x86_64:{}'.format(stress_ng_tag),
                                   'arm64': 'brafsn/stress-ng-arm64:{}'.format(stress_ng_tag)},
                        },
    'stress-ng:matrixprod': {'program': 'stress-ng',
                             'name': 'matrix product of two 128 x 128 matrices of double floats',
                             'cmd': "--cpu {numcpu} --cpu-method matrixprod -t 5 --metrics 2>&1 | tail -1 | awk '{{print $9}}'",
                             'timeout': 10,
                             'images': {'x86_64': 'brafsn/stress-ng-x86_64:{}'.format(stress_ng_tag),
                                        'arm64': 'brafsn/stress-ng-arm64:{}'.format(stress_ng_tag)},
                             }
}



import os
import re
import sys
import signal
import traceback
import click
import pandas as pd
import pytimeparse
import boto3
from cloudperf import get_prices, get_performance, get_combined, prices_url, performance_url, terminate_instances
from cloudperf.core import fail_on_exit

try:
    import faulthandler
    faulthandler.register(signal.SIGUSR1)
except ImportError:
    pass


@click.group()
def main():
    pass


def get_comp(file):
    fn, ext = os.path.splitext(file)
    comp = None
    try:
        ext = ext[1:]
        if ext in ('gzip', 'bz2', 'zip', 'xz'):
            comp = ext
        if ext == 'gz':
            comp = 'gzip'
    except Exception:
        pass
    return comp


def s3_upload(s3_bucket, file):
    comp = get_comp(file)
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(s3_bucket)
    if comp == 'gzip':
        # upload with gzip Content-Encoding and proper Content-Type
        bucket.upload_file(file, os.path.basename(file),
                           ExtraArgs={'ACL': 'public-read',
                                      'ContentType': 'application/json; charset=utf-8',
                                      'ContentEncoding': 'gzip'})
    elif comp:
        bucket.upload_file(file, os.path.basename(
            file), ExtraArgs={'ACL': 'public-read'})
    else:
        bucket.upload_file(file, os.path.basename(
            file), ExtraArgs={'ACL': 'public-read',
                              'ContentType': 'application/json; charset=utf-8'})


def df_filter(df, filters):
    for f in filters:
        m = re.search('(?P<col>[^=<>]+)(?P<op>[=<>]+)(?P<value>.*)', f)
        if not m:
            continue
        try:
            v = float(m.group('value'))
        except Exception:
            v = m.group('value')
        if m.group('op') == '=':
            df = df[df[m.group('col')] == v]
        elif m.group('op') == '>':
            df = df[df[m.group('col')] > v]
        elif m.group('op') == '<':
            df = df[df[m.group('col')] < v]
        elif m.group('op') == '<=':
            df = df[df[m.group('col')] <= v]
        elif m.group('op') == '>=':
            df = df[df[m.group('col')] >= v]
    return df


@main.command()
@click.option('--prices', help='Prices URL (pandas.read_json)', default=prices_url, show_default=True)
@click.option('--file', help='Write prices to this file', required=True)
@click.option('--s3-bucket', help='Write prices to this s3 bucket')
@click.option('--update/--no-update',
              help='Read file first and update it with new data, leaving disappeared entries there for historical reasons',
              default=True, show_default=True)
def write_prices(prices, file, s3_bucket, update):
    if not update:
        prices = None
    df = get_prices(prices, update)
    df.to_json(file, orient='records', compression=get_comp(file), date_unit='s')
    if s3_bucket is not None:
        s3_upload(s3_bucket, file)
    if fail_on_exit():
        sys.exit(1)


@main.command()
@click.option('--prices', help='Prices URL (pandas.read_json)', default=prices_url, show_default=True)
@click.option('--perf', help='Performance URL (pandas.read_json)', default=performance_url, show_default=True)
@click.option('--file', help='Write performance data to this file', required=True)
@click.option('--s3-bucket', help='Write data to this s3 bucket')
@click.option('--update/--no-update',
              help='Read file first and update it with new data, leaving disappeared entries there for historical reasons',
              default=True, show_default=True)
@click.option('--expire', help='Re-run benchmarks after this time', default='12w', show_default=True)
@click.option('--terminate/--no-terminate',
              help='Terminate tagged images at the end of the run to clean up leftover ones',
              default=False, show_default=True)
@click.option('--tag', help='Add these tags to EC2 instances (key:value format)', multiple=True)
def write_performance(prices, perf, file, s3_bucket, update, expire, terminate, tag):
    tags = [i.split(':', 1) for i in tag]
    # convert human readable to seconds
    expire = pytimeparse.parse(expire)
    comp = get_comp(file)
    if not update:
        perf = None
    try:
        get_performance(prices, perf, update, expire, tags=tags).to_json(file, orient='records', compression=comp, date_unit='s')
        if s3_bucket is not None:
            s3_upload(s3_bucket, file)
    except Exception:
        traceback.print_exc()
    finally:
        if terminate:
            terminate_instances()
    if fail_on_exit():
        sys.exit(1)


@main.command()
@click.option('--prices', help='Prices URL (pandas.read_json)', default=prices_url, show_default=True)
@click.option('--perf', help='Performance URL (pandas.read_json)', default=performance_url, show_default=True)
@click.option('--file', help='Write combined perf/price data to this file', required=True)
@click.option('--s3-bucket', help='Write data to this s3 bucket')
def write_combined(prices, perf, file, s3_bucket):
    comp = get_comp(file)
    get_combined(prices, perf).to_json(file, orient='records', compression=comp, date_unit='s')
    if s3_bucket is not None:
        s3_upload(s3_bucket, file)
    if fail_on_exit():
        sys.exit(1)


@main.command()
@click.option('--prices', help='Prices URL (pandas.read_json)', default=prices_url, show_default=True)
@click.option('--cols', help='Columns to show', default=['instanceType', 'region', 'spot-az',
                                                         'vcpu', 'memory', 'price'],
              show_default=True, multiple=True)
@click.option('--sort', help='Sort by these columns', default=['price'], multiple=True, show_default=True)
@click.option('--filter', help="Apply filters like --filter 'benchmark_cpus>4' --filter benchmark_id=sng_zlib", default=[], multiple=True)
def prices(prices, cols, sort, filter):
    df = get_prices(prices)
    df = df_filter(df, filter)
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):
        print(df.sort_values(list(sort))[list(cols)].to_string(index=False))


perf_defcols = ['instanceType', 'benchmark_id', 'benchmark_cpus']


@main.command()
@click.option('--prices', help='Prices URL (pandas.read_json)', default=prices_url, show_default=True)
@click.option('--perf', help='Performance URL (pandas.read_json)', default=performance_url, show_default=True)
@click.option('--cols', help='Columns to show', default=perf_defcols, show_default=True, multiple=True)
@click.option('--sort', help='Sort by these columns', default=['perf/price'], multiple=True, show_default=True)
@click.option('--filter', help="Apply filters like --filter 'benchmark_cpus>4' --filter benchmark_id=sng_zlib", default=[], multiple=True)
@click.option('--combined/--no-combined',
              help='Show combined prices/performance data or just performance',
              default=True, show_default=True)
@click.option('--maxcpu/--no-maxcpu',
              help='Show performance only for maximum number of CPUs',
              default=True, show_default=True)
def performance(prices, perf, cols, sort, filter, combined, maxcpu):
    cols = list(cols)
    if combined:
        df = get_combined(prices, perf, maxcpu)
        if set(cols) == set(perf_defcols):
            # if we're using the default columns, add perf/price and other
            # infos as well
            cols.extend(['perf/price', 'price', 'benchmark_score', 'region', 'spot-az'])
            # keep order and remove duplicates
            seen = {}
            cols = [seen.setdefault(x, x) for x in cols if x not in seen]
    else:
        sort = ['benchmark_score']
        df = get_performance(prices, perf, maxcpu=maxcpu)
        if set(cols) == set(perf_defcols):
            cols.extend(['benchmark_score'])
            seen = {}
            cols = [seen.setdefault(x, x) for x in cols if x not in seen]
    df = df_filter(df, filter)
    with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.float_format', '{:.4f}'.format):
        print(df.sort_values(list(sort))[list(cols)].to_string(index=False))
